#include <stdint.h>
#include <math.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BAUD 38400
#define MYUBRR F_CPU/16/BAUD-1

#define HC_SR04_NUM_DEVICES	4

#define HC_SR04_ECHO_DEV1 PA1
#define HC_SR04_ECHO_DEV2 PA3
#define HC_SR04_ECHO_DEV3 PA5
#define HC_SR04_ECHO_DEV4 PA7

#define HC_SR04_TRIG_DEV1 PA0
#define HC_SR04_TRIG_DEV2 PA2
#define HC_SR04_TRIG_DEV3 PA4
#define HC_SR04_TRIG_DEV4 PA6

#define MIN_SAFE_DISTANCE 10
#define MAX_BEEP_DISTANCE_FRONTAL 15
#define MAX_BEEP_DISTANCE_LATERAL 7

static volatile float saved_distance[HC_SR04_NUM_DEVICES] = { 0 };
static volatile uint8_t flags[HC_SR04_NUM_DEVICES] = { 0 };
static volatile uint8_t timer_count[HC_SR04_NUM_DEVICES] = { 0 };
static volatile uint8_t saved_timer_count[HC_SR04_NUM_DEVICES] = { 0 };
static volatile uint8_t saved_tcnt0[HC_SR04_NUM_DEVICES] = { 0 };
volatile int selected_sensor = 0;


void USART_Init( unsigned int ubrr);
void USART_Transmit( unsigned char data );


void USART_Init( unsigned int ubrr)
{
    UBRR0H = (unsigned char)(ubrr>>8);
    UBRR0L = (unsigned char)ubrr;
    UCSR0B = (1<<TXEN0);
    UCSR0C = (1<<USBS0)|(3<<UCSZ00);
}

void USART_Transmit( unsigned char data )
{
    while ( !( UCSR0A & (1<<UDRE0)) );
    UDR0 = data;
}


ISR(TIMER0_OVF_vect, ISR_NOBLOCK) {
	int i = 0;
	for(i = 0; i < HC_SR04_NUM_DEVICES; i++)
		timer_count[i]++;
}


ISR(PCINT0_vect, ISR_NOBLOCK)
{
	int device = (int)selected_sensor;
	switch(device) {
		case 0:
			if ((PINA & (1 << HC_SR04_ECHO_DEV1)) != 0) {
				TCNT0 = 0;  
				timer_count[device] = 0;
			}
			else {
				saved_tcnt0[device] = TCNT0;
				saved_timer_count[device] = timer_count[device];
				flags[device] = 1;
			}
			break;
	
		case 1:
			if ((PINA & (1 << HC_SR04_ECHO_DEV2)) != 0) {
				TCNT0 = 0;  
				timer_count[device] = 0;
			}
			else {
				saved_tcnt0[device] = TCNT0;
				saved_timer_count[device] = timer_count[device];
				flags[device] = 1;
			}
			break;

		case 2:
			if ((PINA & (1 << HC_SR04_ECHO_DEV3)) != 0) {
				TCNT0 = 0;  
				timer_count[device] = 0;
			}
			else {
				saved_tcnt0[device] = TCNT0;
				saved_timer_count[device] = timer_count[device];
				flags[device] = 1;
			}
			break;

		case 3:
			if ((PINA & (1 << HC_SR04_ECHO_DEV4)) != 0) {
				TCNT0 = 0;  
				timer_count[device] = 0;
			}
			else {
				saved_tcnt0[device] = TCNT0;
				saved_timer_count[device] = timer_count[device];
				flags[device] = 1;
			}
			break;
		default:
			break;
		}

}
void HC_SR04_init()
{
	TCCR0A = 0;
	TCCR0B = 0;
	TIMSK0 = 0;

	sei();
	TCCR0B |= (1 << CS02);
	TIMSK0 |= (1 << TOIE0);

	DDRA |= (1 << HC_SR04_TRIG_DEV1);
	DDRA &= ~(1 << HC_SR04_ECHO_DEV1);
	PORTA &= ~(1 << HC_SR04_TRIG_DEV1);

	DDRA |= (1 << HC_SR04_TRIG_DEV2);
	DDRA &= ~(1 << HC_SR04_ECHO_DEV2);
	PORTA &= ~(1 << HC_SR04_TRIG_DEV2);

	DDRA |= (1 << HC_SR04_TRIG_DEV3);
	DDRA &= ~(1 << HC_SR04_ECHO_DEV3); 
	PORTA &= ~(1 << HC_SR04_TRIG_DEV3);

	DDRA |= (1 << HC_SR04_TRIG_DEV4);
	DDRA &= ~(1 << HC_SR04_ECHO_DEV4);
	PORTA &= ~(1 << HC_SR04_TRIG_DEV4);

	PCICR |= (1 << PCIE0);
	PCMSK0 = 0;
	PCMSK0 |= (1 << PCINT1);
	PCMSK0 |= (1 << PCINT3);
	PCMSK0 |= (1 << PCINT5);
	PCMSK0 |= (1 << PCINT7);
}
float HC_SR04_get_distance(int device)
{
	float sum = 0;
	flags[device] = 0;
	
	selected_sensor = device;

	if(device == 0) {
		PORTA &= ~(1 << HC_SR04_TRIG_DEV1);
		_delay_us(5);
		PORTA |= (1 << HC_SR04_TRIG_DEV1);
		_delay_us(11);
		PORTA &= ~(1 << HC_SR04_TRIG_DEV1);
		while(flags[device] == 0 && timer_count[device] < 5);
		sum = (((int)saved_tcnt0[device] + 255 * saved_timer_count[device]) * 16.0) * 0.017;

	}

	if(device == 1) {
		PORTA &= ~(1 << HC_SR04_TRIG_DEV2);
		_delay_us(5);
		PORTA |= (1 << HC_SR04_TRIG_DEV2);
		_delay_us(11);
		PORTA &= ~(1 << HC_SR04_TRIG_DEV2);
		while(flags[device] == 0 && timer_count[device] < 5);
		sum = (((int)saved_tcnt0[device] + 255 * saved_timer_count[device]) * 16.0) * 0.017;

	}
		
	if(device == 2) {
		PORTA &= ~(1 << HC_SR04_TRIG_DEV3);
		_delay_us(5);
		PORTA |= (1 << HC_SR04_TRIG_DEV3);
		_delay_us(11);
		PORTA &= ~(1 << HC_SR04_TRIG_DEV3);
		while(flags[device] == 0 && timer_count[device] < 5);
		sum = (((int)saved_tcnt0[device] + 255 * saved_timer_count[device]) * 16.0) * 0.017;
	}

	if(device == 3) {
		PORTA &= ~(1 << HC_SR04_TRIG_DEV4);
		_delay_us(5);
		PORTA |= (1 << HC_SR04_TRIG_DEV4);
		_delay_us(11);
		PORTA &= ~(1 << HC_SR04_TRIG_DEV4);
		while(flags[device] == 0 && timer_count[device] < 5);
		sum = (((int)saved_tcnt0[device] + 255 * saved_timer_count[device]) * 16.0) * 0.017;
	}	
	return sum;
}


int main() {
	char buffer[20] = { 0 };
	unsigned int minimum = 200, dist = 0, iter = 0;
	int i = 0, j = 0, flag = 0;
	HC_SR04_init();
	USART_Init(MYUBRR);

	DDRD |= (1 << PD0);
	PORTD &= ~(1 << PD0);

	DDRB |= (1 << PB4);
	PORTB &= (1 << PB4);
	
	while(1)
	{
		minimum = 200;
		flag = 0;
		for (i = 0; i < HC_SR04_NUM_DEVICES; i++)
		{
			iter++;
			saved_distance[i] = HC_SR04_get_distance(i);
			
			_delay_ms(60);
			if (saved_distance[i] <= 1)
				continue;
			
			if(saved_distance[i] < minimum) {
				minimum = saved_distance[i];
			}
		}


		if (saved_distance[0] < MAX_BEEP_DISTANCE_LATERAL || saved_distance[1] < MAX_BEEP_DISTANCE_LATERAL) {
			PORTD |= (1 << PD0);
			iter = 0;
			flag = 1;
		}
		
		if (saved_distance[2] < MAX_BEEP_DISTANCE_FRONTAL || saved_distance[3] < MAX_BEEP_DISTANCE_FRONTAL) {
			PORTD |= (1 << PD0);
			iter = 0;
			flag = 1;
		}

		if (flag) {
			PORTB &= ~(1 << PB4);
		} else {
			PORTB |= (1 << PB4);
		}

		if (iter > 3)
			PORTD &= ~(1 << PD0);

		_delay_ms(10);

	}
	return 0;
}